package sun.security.x509;

import java.security.AccessController;
import sun.security.action.GetBooleanAction;
import java.text.Normalizer;
import java.util.Locale;
import java.io.OutputStream;
import sun.security.util.DerOutputStream;
import sun.security.util.DerInputStream;
import java.util.List;
import sun.security.pkcs.PKCS9Attribute;
import java.util.ArrayList;
import java.io.ByteArrayOutputStream;
import java.util.Collections;
import java.util.Map;
import java.io.IOException;
import java.io.Reader;
import sun.security.util.DerValue;
import sun.security.util.ObjectIdentifier;
import sun.security.util.Debug;
import sun.security.util.DerEncoder;

public class AVA implements DerEncoder
{
    private static final Debug debug;
    private static final boolean PRESERVE_OLD_DC_ENCODING;
    static final int DEFAULT = 1;
    static final int RFC1779 = 2;
    static final int RFC2253 = 3;
    final ObjectIdentifier oid;
    final DerValue value;
    private static final String specialChars1779 = ",=\n+<>#;\\\"";
    private static final String specialChars2253 = ",=+<>#;\\\"";
    private static final String specialCharsDefault = ",=\n+<>#;\\\" ";
    private static final String escapedDefault = ",+<>;\"";
    private static final String hexDigits = "0123456789ABCDEF";
    
    public AVA(ObjectIdentifier oid, DerValue value) {
        if (oid == null || value == null) {
            throw new NullPointerException();
        }
        this.oid = oid;
        this.value = value;
        return;
    }
    
    AVA(Reader reader) throws IOException {
        this(reader, 1);
    }
    
    AVA(Reader reader, Map<String, String> map) throws IOException {
        this(reader, 1, map);
    }
    
    AVA(Reader reader, int n) throws IOException {
        this(reader, n, Collections.emptyMap());
    }
    
    AVA(Reader reader, int n, Map<String, String> map) throws IOException {
        StringBuilder sb = new StringBuilder();
        while (true) {
            int char1 = readChar(reader, "Incorrect AVA format");
            if (char1 == 61) {
                break;
            }
            sb.append((char)char1);
        }
        this.oid = AVAKeyword.getOID(sb.toString(), n, map);
        sb.setLength(0);
        int n2;
        if (n == 3) {
            n2 = reader.read();
            if (n2 == 32) {
                throw new IOException("Incorrect AVA RFC2253 format - leading space must be escaped");
            }
        }
        else {
            do {
                n2 = reader.read();
            } while (n2 == 32 || n2 == 10);
        }
        if (n2 == -1) {
            this.value = new DerValue("");
            return;
        }
        if (n2 == 35) {
            this.value = parseHexString(reader, n);
        }
        else if (n2 == 34 && n != 3) {
            this.value = this.parseQuotedString(reader, sb);
        }
        else {
            this.value = this.parseString(reader, n2, n, sb);
        }
        return;
    }
    
    public ObjectIdentifier getObjectIdentifier() {
        return this.oid;
    }
    
    public DerValue getDerValue() {
        return this.value;
    }
    
    public String getValueString() {
        try {
            String asString = this.value.getAsString();
            if (asString == null) {
                throw new RuntimeException("AVA string is null");
            }
            return asString;
        }
        catch (IOException ex) {
            throw new RuntimeException("AVA error: " + ex, ex);
        }
    }
    
    private static DerValue parseHexString(Reader reader, int n) throws IOException {
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
        int b = 0;
        int n2 = 0;
        while (true) {
            int read = reader.read();
            if (isTerminator(read, n)) {
                if (n2 == 0) {
                    throw new IOException("AVA parse, zero hex digits");
                }
                if (n2 % 2 == 1) {
                    throw new IOException("AVA parse, odd number of hex digits");
                }
                return new DerValue(byteArrayOutputStream.toByteArray());
            }
            else {
                int index = "0123456789ABCDEF".indexOf((int)Character.toUpperCase((char)read));
                if (index == -1) {
                    throw new IOException("AVA parse, invalid hex digit: " + (char)read);
                }
                if (n2 % 2 == 1) {
                    b = (byte)(b * 16 + (byte)index);
                    byteArrayOutputStream.write(b);
                }
                else {
                    b = (byte)index;
                }
                ++n2;
            }
        }
    }
    
    private DerValue parseQuotedString(Reader reader, StringBuilder sb) throws IOException {
        int i = readChar(reader, "Quoted string did not end in quote");
        ArrayList<Byte> list = new ArrayList();
        boolean n = true;
        while (i != 34) {
            if (i == 92) {
                i = readChar(reader, "Quoted string did not end in quote");
                Byte embeddedHexPair;
                if ((embeddedHexPair = getEmbeddedHexPair(i, reader)) != null) {
                    n = false;
                    list.add(embeddedHexPair);
                    i = reader.read();
                    continue;
                }
                if (",=\n+<>#;\\\"".indexOf((int)(char)i) < 0) {
                    throw new IOException("Invalid escaped character in AVA: " + (char)i);
                }
            }
            if (list.size() > 0) {
                sb.append(getEmbeddedHexString(list));
                list.clear();
            }
            n &= DerValue.isPrintableStringChar((char)i);
            sb.append((char)i);
            i = readChar(reader, "Quoted string did not end in quote");
        }
        if (list.size() > 0) {
            sb.append(getEmbeddedHexString(list));
            list.clear();
        }
        int read;
        do {
            read = reader.read();
        } while (read == 10 || read == 32);
        if (read != -1) {
            throw new IOException("AVA had characters other than whitespace after terminating quote");
        }
        if (this.oid.equals(PKCS9Attribute.EMAIL_ADDRESS_OID) || (this.oid.equals(X500Name.DOMAIN_COMPONENT_OID) && !AVA.PRESERVE_OLD_DC_ENCODING)) {
            return new DerValue((byte)22, sb.toString().trim());
        }
        if (!(n == 0)) {
            return new DerValue(sb.toString().trim());
        }
        return new DerValue((byte)12, sb.toString().trim());
    }
    
    private DerValue parseString(Reader reader, int n, int n2, StringBuilder sb) throws IOException {
        ArrayList<Byte> list = new ArrayList();
        boolean n3 = true;
        int n4 = 1;
        int n5 = 0;
        do {
            boolean n6 = false;
            if (n == 92) {
                n6 = true;
                n = readChar(reader, "Invalid trailing backslash");
                Byte embeddedHexPair;
                if ((embeddedHexPair = getEmbeddedHexPair(n, reader)) != null) {
                    n3 = false;
                    list.add(embeddedHexPair);
                    n = reader.read();
                    n4 = 0;
                    continue;
                }
                if (n2 == 1 && ",=\n+<>#;\\\" ".indexOf((int)(char)n) == -1) {
                    throw new IOException("Invalid escaped character in AVA: '" + (char)n + "'");
                }
                if (n2 == 3) {
                    if (n == 32) {
                        if (n4 == 0 && !trailingSpace(reader)) {
                            throw new IOException("Invalid escaped space character in AVA.  Only a leading or trailing space character can be escaped.");
                        }
                    }
                    else if (n == 35) {
                        if (n4 == 0) {
                            throw new IOException("Invalid escaped '#' character in AVA.  Only a leading '#' can be escaped.");
                        }
                    }
                    else if (",=+<>#;\\\"".indexOf((int)(char)n) == -1) {
                        throw new IOException("Invalid escaped character in AVA: '" + (char)n + "'");
                    }
                }
            }
            else if (n2 == 3) {
                if (",=+<>#;\\\"".indexOf((int)(char)n) != -1) {
                    throw new IOException("Character '" + (char)n + "' in AVA appears without escape");
                }
            }
            else if (",+<>;\"".indexOf((int)(char)n) != -1) {
                throw new IOException("Character '" + (char)n + "' in AVA appears without escape");
            }
            if (list.size() > 0) {
                for (int i = 0; i < n5; ++i) {
                    sb.append(" ");
                }
                n5 = 0;
                sb.append(getEmbeddedHexString(list));
                list.clear();
            }
            n3 &= DerValue.isPrintableStringChar((char)n);
            if (n == 32 && n6 == 0) {
                ++n5;
            }
            else {
                for (int j = 0; j < n5; ++j) {
                    sb.append(" ");
                }
                n5 = 0;
                sb.append((char)n);
            }
            n = reader.read();
            n4 = 0;
        } while (!isTerminator(n, n2));
        if (n2 == 3 && n5 > 0) {
            throw new IOException("Incorrect AVA RFC2253 format - trailing space must be escaped");
        }
        if (list.size() > 0) {
            sb.append(getEmbeddedHexString(list));
            list.clear();
        }
        if (this.oid.equals(PKCS9Attribute.EMAIL_ADDRESS_OID) || (this.oid.equals(X500Name.DOMAIN_COMPONENT_OID) && !AVA.PRESERVE_OLD_DC_ENCODING)) {
            return new DerValue((byte)22, sb.toString());
        }
        if (!(n3 == 0)) {
            return new DerValue(sb.toString());
        }
        return new DerValue((byte)12, sb.toString());
    }
    
    private static Byte getEmbeddedHexPair(int n, Reader reader) throws IOException {
        if ("0123456789ABCDEF".indexOf((int)Character.toUpperCase((char)n)) >= 0) {
            int char1 = readChar(reader, "unexpected EOF - escaped hex value must include two valid digits");
            if ("0123456789ABCDEF".indexOf((int)Character.toUpperCase((char)char1)) >= 0) {
                return new Byte((byte)((Character.digit((char)n, 16) << 4) + Character.digit((char)char1, 16)));
            }
            throw new IOException("escaped hex value must include two valid digits");
        }
        else {
            return null;
        }
    }
    
    private static String getEmbeddedHexString(List<Byte> list) throws IOException {
        int size = list.size();
        byte[] bytes = new byte[size];
        for (int i = 0; i < size; ++i) {
            bytes[i] = (byte)list.get(i);
        }
        return new String(bytes, "UTF8");
    }
    
    private static boolean isTerminator(int n, int n2) {
        switch (n) {
            case -1:
            case 43:
            case 44: {
                return true;
            }
            case 59: {
                return n2 != 3;
            }
            default: {
                return false;
            }
        }
    }
    
    private static int readChar(Reader reader, String message) throws IOException {
        int read = reader.read();
        if (read == -1) {
            throw new IOException(message);
        }
        return read;
    }
    
    private static boolean trailingSpace(Reader reader) throws IOException {
        if (!reader.markSupported()) {
            return true;
        }
        reader.mark(9999);
        int n;
        while (true) {
            int read = reader.read();
            if (read == -1) {
                n = true;
                break;
            }
            if (read == 32) {
                continue;
            }
            if (read == 92) {
                if (reader.read() != 32) {
                    n = false;
                    break;
                }
                continue;
            }
            else {
                n = false;
                break;
            }
        }
        reader.reset();
        return n;
    }
    
    AVA(DerValue derValue) throws IOException {
        if (derValue.tag != 48) {
            throw new IOException("AVA not a sequence");
        }
        this.oid = derValue.data.getOID();
        this.value = derValue.data.getDerValue();
        if (derValue.data.available() != 0) {
            throw new IOException("AVA, extra bytes = " + derValue.data.available());
        }
        return;
    }
    
    AVA(DerInputStream derInputStream) throws IOException {
        this(derInputStream.getDerValue());
    }
    
    @Override
    public boolean equals(Object o) {
        return this == o || (o instanceof AVA && this.toRFC2253CanonicalString().equals(((AVA)o).toRFC2253CanonicalString()));
    }
    
    @Override
    public int hashCode() {
        return this.toRFC2253CanonicalString().hashCode();
    }
    
    @Override
    public void encode(DerOutputStream derOutputStream) throws IOException {
        this.derEncode(derOutputStream);
    }
    
    public void derEncode(OutputStream outputStream) throws IOException {
        DerOutputStream out = new DerOutputStream();
        DerOutputStream derOutputStream = new DerOutputStream();
        out.putOID(this.oid);
        this.value.encode(out);
        derOutputStream.write((byte)48, out);
        outputStream.write(derOutputStream.toByteArray());
    }
    
    private String toKeyword(int n, Map<String, String> map) {
        return AVAKeyword.getKeyword(this.oid, n, map);
    }
    
    @Override
    public String toString() {
        return this.toKeywordValueString(this.toKeyword(1, Collections.emptyMap()));
    }
    
    public String toRFC1779String() {
        return this.toRFC1779String(Collections.emptyMap());
    }
    
    public String toRFC1779String(Map<String, String> map) {
        return this.toKeywordValueString(this.toKeyword(2, map));
    }
    
    public String toRFC2253String() {
        return this.toRFC2253String(Collections.emptyMap());
    }
    
    public String toRFC2253String(Map<String, String> map) {
        StringBuilder sb = new StringBuilder(100);
        sb.append(this.toKeyword(3, map));
        sb.append('=');
        if ((sb.charAt(0) >= '0' && sb.charAt(0) <= '9') || !isDerString(this.value, false)) {
            byte[] byteArray;
            try {
                byteArray = this.value.toByteArray();
            }
            catch (IOException ex3) {
                throw new IllegalArgumentException("DER Value conversion");
            }
            sb.append('#');
            for (int l = 0; l < byteArray.length; ++l) {
                int n3 = byteArray[l];
                sb.append(Character.forDigit(0xF & n3 >>> 4, 16));
                sb.append(Character.forDigit(0xF & n3, 16));
            }
        }
        else {
            String s;
            try {
                s = new String(this.value.getDataBytes(), "UTF8");
            }
            catch (IOException ex) {
                throw new IllegalArgumentException("DER Value conversion");
            }
            StringBuilder sb2 = new StringBuilder();
            for (int i = 0; i < s.length(); ++i) {
                int char1 = s.charAt(i);
                if (DerValue.isPrintableStringChar((char)char1) || ",=+<>#;\"\\".indexOf(char1) >= 0) {
                    if (",=+<>#;\"\\".indexOf(char1) >= 0) {
                        sb2.append('\\');
                    }
                    sb2.append((char)char1);
                }
                else if (char1 == '\0') {
                    sb2.append("\\00");
                }
                else if (AVA.debug != null && Debug.isOn("ava")) {
                    byte[] bytes;
                    try {
                        bytes = Character.toString((char)char1).getBytes("UTF8");
                    }
                    catch (IOException ex2) {
                        throw new IllegalArgumentException("DER Value conversion");
                    }
                    for (int j = 0; j < bytes.length; ++j) {
                        sb2.append('\\');
                        sb2.append(Character.toUpperCase(Character.forDigit(0xF & bytes[j] >>> 4, 16)));
                        sb2.append(Character.toUpperCase(Character.forDigit(0xF & bytes[j], 16)));
                    }
                }
                else {
                    sb2.append((char)char1);
                }
            }
            char[] charArray = sb2.toString().toCharArray();
            StringBuilder o = new StringBuilder();
            int n;
            for (n = 0; n < charArray.length && (charArray[n] == ' ' || charArray[n] == '\r'); ++n) {}
            int n2;
            for (n2 = charArray.length - 1; n2 >= 0 && (charArray[n2] == ' ' || charArray[n2] == '\r'); --n2) {}
            for (int k = 0; k < charArray.length; ++k) {
                int c = charArray[k];
                if (k < n || k > n2) {
                    o.append('\\');
                }
                o.append((char)c);
            }
            sb.append(o.toString());
        }
        return sb.toString();
    }
    
    public String toRFC2253CanonicalString() {
        StringBuilder sb = new StringBuilder(40);
        sb.append(this.toKeyword(3, Collections.emptyMap()));
        sb.append('=');
        if ((sb.charAt(0) >= '0' && sb.charAt(0) <= '9') || !isDerString(this.value, true)) {
            byte[] byteArray;
            try {
                byteArray = this.value.toByteArray();
            }
            catch (IOException ex3) {
                throw new IllegalArgumentException("DER Value conversion");
            }
            sb.append('#');
            for (int k = 0; k < byteArray.length; ++k) {
                int n2 = byteArray[k];
                sb.append(Character.forDigit(0xF & n2 >>> 4, 16));
                sb.append(Character.forDigit(0xF & n2, 16));
            }
        }
        else {
            String s;
            try {
                s = new String(this.value.getDataBytes(), "UTF8");
            }
            catch (IOException ex) {
                throw new IllegalArgumentException("DER Value conversion");
            }
            StringBuilder sb2 = new StringBuilder();
            int n = 0;
            for (int i = 0; i < s.length(); ++i) {
                int char1 = s.charAt(i);
                if (DerValue.isPrintableStringChar((char)char1) || ",+<>;\"\\".indexOf(char1) >= 0 || (i == 0 && char1 == '#')) {
                    if ((i == 0 && char1 == '#') || ",+<>;\"\\".indexOf(char1) >= 0) {
                        sb2.append('\\');
                    }
                    if (!Character.isWhitespace((char)char1)) {
                        n = 0;
                        sb2.append((char)char1);
                    }
                    else if (n == 0) {
                        n = 1;
                        sb2.append((char)char1);
                    }
                }
                else if (AVA.debug != null && Debug.isOn("ava")) {
                    n = 0;
                    byte[] bytes;
                    try {
                        bytes = Character.toString((char)char1).getBytes("UTF8");
                    }
                    catch (IOException ex2) {
                        throw new IllegalArgumentException("DER Value conversion");
                    }
                    for (int j = 0; j < bytes.length; ++j) {
                        sb2.append('\\');
                        sb2.append(Character.forDigit(0xF & bytes[j] >>> 4, 16));
                        sb2.append(Character.forDigit(0xF & bytes[j], 16));
                    }
                }
                else {
                    n = 0;
                    sb2.append((char)char1);
                }
            }
            sb.append(sb2.toString().trim());
        }
        return Normalizer.normalize(sb.toString().toUpperCase(Locale.US).toLowerCase(Locale.US), Normalizer.Form.NFKD);
    }
    
    private static boolean isDerString(DerValue derValue, boolean b) {
        if (b) {
            switch (derValue.tag) {
                case 12:
                case 19: {
                    return true;
                }
                default: {
                    return false;
                }
            }
        }
        else {
            switch (derValue.tag) {
                case 12:
                case 19:
                case 20:
                case 22:
                case 27:
                case 30: {
                    return true;
                }
                default: {
                    return false;
                }
            }
        }
    }
    
    boolean hasRFC2253Keyword() {
        return AVAKeyword.hasKeyword(this.oid, 3);
    }
    
    private String toKeywordValueString(String str) {
        StringBuilder sb = new StringBuilder(40);
        sb.append(str);
        sb.append("=");
        try {
            String asString = this.value.getAsString();
            if (asString == null) {
                byte[] byteArray = this.value.toByteArray();
                sb.append('#');
                for (int k = 0; k < byteArray.length; ++k) {
                    sb.append("0123456789ABCDEF".charAt(byteArray[k] >> 4 & 0xF));
                    sb.append("0123456789ABCDEF".charAt(byteArray[k] & 0xF));
                }
            }
            else {
                int n = 0;
                StringBuilder sb2 = new StringBuilder();
                int n2 = 0;
                int length = asString.length();
                int n3 = length > 1 && asString.charAt(0) == '\"' && asString.charAt(length - 1) == '\"';
                for (int i = 0; i < length; ++i) {
                    int char1 = asString.charAt(i);
                    if (n3 && (i == 0 || i == length - 1)) {
                        sb2.append((char)char1);
                    }
                    else if (DerValue.isPrintableStringChar((char)char1) || ",+=\n<>#;\\\"".indexOf(char1) >= 0) {
                        if (n == 0 && ((i == 0 && (char1 == ' ' || char1 == '\n')) || ",+=\n<>#;\\\"".indexOf(char1) >= 0)) {
                            n = 1;
                        }
                        if (char1 != ' ' && char1 != '\n') {
                            if (char1 == '\"' || char1 == '\\') {
                                sb2.append('\\');
                            }
                            n2 = 0;
                        }
                        else {
                            if (n == 0 && n2 != 0) {
                                n = 1;
                            }
                            n2 = 1;
                        }
                        sb2.append((char)char1);
                    }
                    else if (AVA.debug != null && Debug.isOn("ava")) {
                        n2 = 0;
                        byte[] bytes = Character.toString((char)char1).getBytes("UTF8");
                        for (int j = 0; j < bytes.length; ++j) {
                            sb2.append('\\');
                            sb2.append(Character.toUpperCase(Character.forDigit(0xF & bytes[j] >>> 4, 16)));
                            sb2.append(Character.toUpperCase(Character.forDigit(0xF & bytes[j], 16)));
                        }
                    }
                    else {
                        n2 = 0;
                        sb2.append((char)char1);
                    }
                }
                if (sb2.length() > 0) {
                    int char2 = sb2.charAt(sb2.length() - 1);
                    if (char2 == ' ' || char2 == '\n') {
                        n = 1;
                    }
                }
                if (n3 == 0 && n != 0) {
                    sb.append("\"" + sb2.toString() + "\"");
                }
                else {
                    sb.append(sb2.toString());
                }
            }
        }
        catch (IOException ex) {
            throw new IllegalArgumentException("DER Value conversion");
        }
        return sb.toString();
    }
    
    static {
        debug = Debug.getInstance("x509", "\t[AVA]");
        PRESERVE_OLD_DC_ENCODING = (boolean)AccessController.doPrivileged(new GetBooleanAction("com.sun.security.preserveOldDCEncoding"));
    }
}
